Package: Infusion
Type: Package
Title: Inference Using Simulation
Description: Implements functions for simulation-based inference. In particular, implements functions to perform likelihood inference from data summaries whose distributions are simulated, as described in Rousset et al. (2026) <doi:10.24072/pcjournal.721>.  
Encoding: UTF-8
Version: 2.3.12
Date: 2026-05-11
Imports: spaMM (>= 4.5.0), proxy, blackbox (>= 1.1.41), mvtnorm,
        methods, numDeriv, pbapply, ranger, foreach, matrixStats, boot,
        nloptr, geometry, cli, viridisLite
Suggests: testthat, Rmixmod, caret, xLLiM, reticulate, mafR (>= 1.1.9)
Depends: R (>= 3.3.0)
Authors@R: c(person("François", "Rousset", role = c("aut", "cre", "cph"),
                     email = "francois.rousset@umontpellier.fr",
                     comment = c(ORCID = "0000-0003-4670-0371")))
Maintainer: François Rousset <francois.rousset@umontpellier.fr>
License: CeCILL-2
ByteCompile: true
URL: https://gitlab.mbb.univ-montp2.fr/francois/Infusion
NeedsCompilation: no
Packaged: 2026-05-11 10:57:24 UTC; francois.rousset
Author: François Rousset [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0003-4670-0371>)
Repository: CRAN
Date/Publication: 2026-05-11 15:40:02 UTC
Built: R 4.5.2; ; 2026-05-11 15:48:23 UTC; unix
